package org.commonjava.tensor.agg;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import javax.enterprise.context.ApplicationScoped;
import javax.inject.Inject;

import org.apache.maven.graph.common.ref.ProjectVersionRef;
import org.apache.maven.graph.common.version.InvalidVersionSpecificationException;
import org.apache.maven.graph.effective.EProjectCycle;
import org.apache.maven.graph.effective.EProjectCycle.Builder;
import org.apache.maven.graph.effective.EProjectGraph;
import org.apache.maven.graph.effective.filter.AnyFilter;
import org.apache.maven.graph.effective.filter.ProjectRelationshipFilter;
import org.apache.maven.graph.effective.ref.EProjectKey;
import org.apache.maven.graph.effective.rel.ProjectRelationship;
import org.apache.maven.graph.effective.transform.FilteringGraphTransformer;
import org.apache.maven.graph.spi.GraphDriverException;
import org.commonjava.tensor.data.TensorDataException;
import org.commonjava.tensor.data.TensorDataManager;
import org.commonjava.tensor.discover.DiscoveryConfig;
import org.commonjava.tensor.discover.ProjectRelationshipDiscoverer;
import org.commonjava.tensor.event.TensorEventFunnel;
import org.commonjava.util.logging.Logger;

@ApplicationScoped
public class DefaultGraphAggregator
    implements GraphAggregator
{

    private final Logger logger = new Logger( getClass() );

    @Inject
    private TensorDataManager dataManager;

    @Inject
    private ProjectRelationshipDiscoverer discoverer;

    @Inject
    private TensorEventFunnel funnel;

    protected DefaultGraphAggregator()
    {
    }

    public DefaultGraphAggregator( final TensorDataManager dataManager, final ProjectRelationshipDiscoverer discoverer,
                                   final TensorEventFunnel funnel )
    {
        this.dataManager = dataManager;
        this.discoverer = discoverer;
        this.funnel = funnel;
    }

    /* (non-Javadoc)
     * @see org.commonjava.tensor.agg.GraphCrawler#connectIncompleteSubgraphs(org.apache.maven.graph.effective.EProjectGraph, org.apache.maven.graph.effective.filter.ProjectRelationshipFilter, org.commonjava.tensor.agg.DiscoveryConfig, boolean)
     */
    @Override
    public EProjectGraph connectSubgraphs( final EProjectGraph graph, final AggregationOptions config )
        throws TensorDataException
    {
        final ProjectRelationshipFilter filter = config.getFilter();
        if ( graph != null )
        {
            if ( config.isDiscoveryEnabled() )
            {
                ProjectRelationshipFilter f = filter;
                if ( f == null )
                {
                    f = new AnyFilter();
                }

                final LinkedList<EProjectGraph> graphs = new LinkedList<EProjectGraph>();
                graphs.add( graph );

                int changeCount = 0;
                do
                {
                    changeCount =
                        connectSubgraphs( graph, graphs, new HashSet<EProjectCycle.Builder>(), config, f,
                                          new HashSet<ProjectVersionRef>(), new HashSet<ProjectVersionRef>() );

                    logger.info( "%d changes...re-running to see if we can discover a bit more." );
                }
                while ( changeCount > 0 );
            }

            if ( filter != null )
            {
                // if we have a filter, it's still possible that we loaded a graph
                // from storage that contains more than what we would have filtered
                // out if we had to reconnect it, above. In this case, we would want
                // to do a second-level filter to make sure stored relationships
                // were subjected to the same filter.
                final FilteringGraphTransformer txform = new FilteringGraphTransformer( filter );
                try
                {
                    graph.traverse( txform );
                }
                catch ( final GraphDriverException e )
                {
                    throw new TensorDataException( "Failed to traverse: %s. Reason: %s", e, graph.getRoot(),
                                                   e.getMessage() );
                }

                return (EProjectGraph) txform.getTransformedNetwork();
            }
            else
            {
                return graph;
            }
        }

        return graph;
    }

    private int connectSubgraphs( final EProjectGraph graph, final LinkedList<EProjectGraph> graphs,
                                  final Set<EProjectCycle.Builder> cycleBuilders, final AggregationOptions config,
                                  final ProjectRelationshipFilter filter,
                                  final Set<ProjectVersionRef> cycleParticipants, final Set<ProjectVersionRef> missing )
        throws TensorDataException
    {
        //        logger.info( "Reconnecting: %s using filter: %s", graph.getRoot(), filter );
        final long timeoutMillis = config.getDiscoveryTimeoutMillis();
        final DiscoveryConfig discoveryConfig = config.getDiscoveryConfig();

        int changeCount = 0;
        for ( final ProjectVersionRef ref : new HashSet<ProjectVersionRef>( graph.getIncompleteSubgraphs() ) )
        {
            logger.info( "\n\n\n\nProcessing missing project: %s\n\n\n\n", ref );
            //            logger.info( "%s marked as missing in: %s", ref, graph.getRoot() );
            final Set<EProjectCycle> cycles = graph.getCycles();
            if ( cycles != null )
            {
                for ( final EProjectCycle cycle : cycles )
                {
                    final Set<ProjectVersionRef> participants = cycle.getAllParticipatingProjects();
                    cycleParticipants.addAll( participants );
                }
            }

            if ( cycleParticipants.contains( ref ) )
            {
                //                logger.warn( "Dependency-cycle participant detected: %s. It was marked incomplete in graph: %s. Skipping reconnection.",
                //                             ref, graph.getRoot() );
                continue;
            }

            final Set<ProjectRelationship<?>> rels = graph.getRelationshipsTargeting( ref );
            if ( detectCycles( ref, rels, graphs, cycleBuilders, filter, cycleParticipants ) )
            {
                break;
            }

            if ( discoverer != null && !missing.contains( ref ) )
            {
                try
                {
                    ref.getVersionSpec();

                    // kick off the async process of resolving metadata and building the graph.
                    // then, wait for the event funnel to capture something telling us to re-check
                    // for the project graph.
                    discoverer.discoverRelationships( ref, discoveryConfig );
                    funnel.waitForGraph( ref, timeoutMillis );

                    if ( !graph.connectFor( new EProjectKey( ref ) ) )
                    {
                        missing.add( ref );
                    }
                    else
                    {
                        changeCount++;
                    }
                }
                catch ( final InvalidVersionSpecificationException e )
                {
                    logger.error( "Cannot discover subgraph for: %s. Reason: %s.", e, ref, e.getMessage() );
                }
                catch ( final GraphDriverException e )
                {
                    throw new TensorDataException( "Failed to connect graph for: %s to our graph: %s. Reason: %s", e,
                                                   ref, graph.getRoot(), e.getMessage() );
                }
            }
        }

        return changeCount;
    }

    private boolean detectCycles( final ProjectVersionRef ref, final Set<ProjectRelationship<?>> rels,
                                  final List<EProjectGraph> graphs, final Set<Builder> cycleBuilders,
                                  final ProjectRelationshipFilter filters,
                                  final Set<ProjectVersionRef> cycleParticipants )
    {
        boolean foundCycle = false;

        for ( final EProjectCycle.Builder cycleBuilder : cycleBuilders )
        {
            final int idx = cycleBuilder.indexOf( ref );
            if ( idx > -1 )
            {
                // CYCLE!!!
                final EProjectCycle.Builder cb = new EProjectCycle.Builder( cycleBuilder, idx );
                for ( final ProjectRelationship<?> rel : rels )
                {
                    cycleParticipants.add( rel.getDeclaring() );
                    cycleParticipants.add( rel.getTarget()
                                              .asProjectVersionRef() );

                    if ( !filters.accept( rel ) )
                    {
                        continue;
                    }

                    cb.with( rel );

                    final EProjectCycle cycle = cb.build();
                    //                    logger.info( "CYCLE! %s", cycle );

                    for ( final EProjectGraph g : graphs )
                    {
                        g.addCycle( cycle );
                    }

                    cb.withoutLast();
                    foundCycle = true;
                }
            }
        }

        return foundCycle;
    }

}
