/*******************************************************************************
 * Copyright 2011 John Casey
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.commonjava.tensor.data;

import static org.apache.commons.lang.StringUtils.isEmpty;
import static org.apache.commons.lang.StringUtils.join;
import static org.apache.maven.graph.effective.util.EGraphUtils.filter;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.enterprise.context.ApplicationScoped;
import javax.enterprise.event.Event;
import javax.inject.Inject;

import org.apache.maven.graph.common.ref.ArtifactRef;
import org.apache.maven.graph.common.ref.ProjectRef;
import org.apache.maven.graph.common.ref.ProjectVersionRef;
import org.apache.maven.graph.effective.EProjectGraph;
import org.apache.maven.graph.effective.EProjectRelationships;
import org.apache.maven.graph.effective.EProjectWeb;
import org.apache.maven.graph.effective.filter.ParentFilter;
import org.apache.maven.graph.effective.filter.ProjectRelationshipFilter;
import org.apache.maven.graph.effective.ref.EProjectKey;
import org.apache.maven.graph.effective.rel.ProjectRelationship;
import org.apache.maven.graph.effective.traverse.AncestryTraversal;
import org.apache.maven.graph.spi.GraphDriverException;
import org.commonjava.tensor.event.MissingRelationshipsEvent;
import org.commonjava.tensor.event.NewRelationshipsEvent;
import org.commonjava.util.logging.Logger;

@ApplicationScoped
public class DefaultTensorDataManager
    implements TensorDataManager
{

    private final Logger logger = new Logger( getClass() );

    @Inject
    private EProjectWeb globalWeb;

    @Inject
    private Event<NewRelationshipsEvent> newRels;

    @Inject
    private Event<MissingRelationshipsEvent> missingRels;

    public DefaultTensorDataManager()
    {
    }

    public DefaultTensorDataManager( final EProjectWeb globalWeb )
    {
        this.globalWeb = globalWeb;
    }

    /* (non-Javadoc)
     * @see org.commonjava.tensor.data.TensorDataManager#storeGraph(boolean, boolean, org.apache.maven.graph.effective.EProjectGraph)
     */
    @Override
    public Set<ProjectRelationship<?>> storeGraph( final EProjectGraph graph )
        throws TensorDataException
    {
        if ( graph.isDerivedFrom( globalWeb ) )
        {
            return Collections.emptySet();
        }

        logger.info( "Retrieving all relationships in graph: %s", graph.getRoot() );
        final Set<ProjectRelationship<?>> rels = globalWeb.addAll( graph.getExactAllRelationships() );
        fireMissingEvents( rels );

        return rels;
    }

    /* (non-Javadoc)
     * @see org.commonjava.tensor.data.TensorDataManager#storeRelationships(boolean, org.apache.maven.graph.effective.EProjectRelationships)
     */
    @Override
    public Set<ProjectRelationship<?>> storeRelationships( final EProjectRelationships relationships )
        throws TensorDataException
    {
        final Set<ProjectRelationship<?>> rels = globalWeb.addAll( relationships.getExactAllRelationships() );
        fireMissingEvents( rels );

        return rels;
    }

    /* (non-Javadoc)
     * @see org.commonjava.tensor.data.TensorDataManager#storeRelationships(org.apache.maven.graph.effective.rel.ProjectRelationship)
     */
    @Override
    public Set<ProjectRelationship<?>> storeRelationships( final ProjectRelationship<?>... relationships )
        throws TensorDataException
    {
        final Set<ProjectRelationship<?>> rels = globalWeb.addAll( relationships );
        fireMissingEvents( rels );

        return rels;
    }

    /* (non-Javadoc)
     * @see org.commonjava.tensor.data.TensorDataManager#storeRelationships(java.util.Collection)
     */
    @Override
    public Set<ProjectRelationship<?>> storeRelationships( final Collection<ProjectRelationship<?>> relationships )
        throws TensorDataException
    {
        final Set<ProjectRelationship<?>> rels = globalWeb.addAll( relationships );
        fireMissingEvents( rels );

        return rels;
    }

    private void fireMissingEvents( final Set<ProjectRelationship<?>> stored )
    {
        if ( missingRels == null )
        {
            return;
        }

        for ( final ProjectRelationship<?> rel : stored )
        {
            ProjectVersionRef ref = rel.getTarget();
            if ( ref instanceof ArtifactRef )
            {
                ref = ( (ArtifactRef) ref ).asProjectVersionRef();
            }

            if ( !globalWeb.containsGraph( new EProjectKey( ref ) ) )
            {
                missingRels.fire( new MissingRelationshipsEvent( ref ) );
            }
        }
    }

    @Override
    public EProjectGraph getProjectGraph( final EProjectKey key )
        throws TensorDataException
    {
        try
        {
            return globalWeb.getGraph( key );
        }
        catch ( final GraphDriverException e )
        {
            throw new TensorDataException( "Failed to retrieve graph for: %s from global graph database. Reason: %s",
                                           e, key.getProject(), e.getMessage() );
        }
    }

    @Override
    public synchronized EProjectGraph getProjectGraph( final ProjectVersionRef ref )
        throws TensorDataException
    {
        return getProjectGraph( new EProjectKey( ref ) );
    }

    /* (non-Javadoc)
     * @see org.commonjava.tensor.data.TensorDataManager#getAncestry(org.apache.maven.graph.common.ref.ProjectVersionRef)
     */
    @Override
    public List<ProjectVersionRef> getAncestry( final ProjectVersionRef source )
        throws TensorDataException
    {
        // This is probably over-optimization, especially now that we're not
        // constructing/parsing queries to retrieve parents.

        // final EProjectGraph graph = getProjectGraph( source, true );
        // if ( graph != null && graph.isComplete() )
        // {
        // final AncestryTraversal traversal = new AncestryTraversal( source );
        // graph.traverse( traversal );
        //
        // final List<ProjectVersionRef> ancestry = traversal.getAncestry();
        // if ( !graph.getIncompleteSubgraphs()
        // .contains( ancestry.get( ancestry.size() - 1 ) ) )
        // {
        // return ancestry;
        // }
        // }
        //

        // FIXME: Need to use a direct cypher query here, this will probably be massively inefficient.
        final AncestryTraversal traversal = new AncestryTraversal();
        try
        {
            globalWeb.traverse( source, traversal );
        }
        catch ( final GraphDriverException e )
        {
            throw new TensorDataException( "Failed to traverse database for parents of: %s. Reason: %s", e, source,
                                           e.getMessage() );
        }

        final List<ProjectVersionRef> ancestry = traversal.getCapturedProjects( true );

        return ancestry.size() == 1 ? null : ancestry;
    }

    /* (non-Javadoc)
     * @see org.commonjava.tensor.data.TensorDataManager#getParent(org.apache.maven.graph.common.ref.ProjectVersionRef)
     */
    @Override
    public ProjectVersionRef getParent( final ProjectVersionRef source )
        throws TensorDataException
    {
        final Set<ProjectRelationship<?>> rels =
            new HashSet<ProjectRelationship<?>>( globalWeb.getDirectRelationships( source ) );
        filter( rels, new ParentFilter( false ) );

        if ( rels.size() > 1 )
        {
            final Set<ProjectVersionRef> refs = new HashSet<ProjectVersionRef>();
            for ( final ProjectRelationship<?> rel : rels )
            {
                refs.add( rel.getDeclaring() );
            }

            throw new TensorDataException( "Found %d parents for %s: %s", rels.size(), source, join( refs, ", " ) );
        }

        return rels.iterator()
                   .next()
                   .getTarget()
                   .asProjectVersionRef();
    }

    /* (non-Javadoc)
     * @see org.commonjava.tensor.data.TensorDataManager#getKnownChildren(org.apache.maven.graph.common.ref.ProjectVersionRef)
     */
    @Override
    public Set<ProjectVersionRef> getKnownChildren( final ProjectVersionRef parent )
        throws TensorDataException
    {
        final Set<ProjectRelationship<?>> rels =
            new HashSet<ProjectRelationship<?>>( globalWeb.getRelationshipsTargeting( parent ) );

        filter( rels, new ParentFilter( false ) );

        final Set<ProjectVersionRef> refs = new HashSet<ProjectVersionRef>();
        for ( final ProjectRelationship<?> rel : rels )
        {
            refs.add( rel.getDeclaring() );
        }

        return refs;
    }

    /* (non-Javadoc)
     * @see org.commonjava.tensor.data.TensorDataManager#getAllDirectRelationshipsWithExactSource(org.apache.maven.graph.common.ref.ProjectVersionRef, org.apache.maven.graph.common.RelationshipType)
     */
    @Override
    public Set<ProjectRelationship<?>> getAllDirectRelationshipsWithExactSource( final ProjectVersionRef source,
                                                                                 final ProjectRelationshipFilter filter )
        throws TensorDataException
    {
        final Set<ProjectRelationship<?>> rels =
            new HashSet<ProjectRelationship<?>>( globalWeb.getDirectRelationships( source ) );

        filter( rels, filter );

        return rels;
    }

    /* (non-Javadoc)
     * @see org.commonjava.tensor.data.TensorDataManager#getAllDirectRelationshipsWithExactTarget(org.apache.maven.graph.common.ref.ProjectVersionRef, org.apache.maven.graph.common.RelationshipType)
     */
    @Override
    public Set<ProjectRelationship<?>> getAllDirectRelationshipsWithExactTarget( final ProjectVersionRef target,
                                                                                 final ProjectRelationshipFilter filter )
        throws TensorDataException
    {
        final Set<ProjectRelationship<?>> rels =
            new HashSet<ProjectRelationship<?>>( globalWeb.getRelationshipsTargeting( target.asProjectVersionRef() ) );
        filter( rels, filter );

        return rels;
    }

    /* (non-Javadoc)
     * @see org.commonjava.tensor.data.TensorDataManager#getAllDirectRelationshipsWithGATarget(org.apache.maven.graph.common.ref.ProjectRef, org.apache.maven.graph.common.RelationshipType)
     */
    @Override
    public Set<ProjectRelationship<?>> getAllDirectRelationshipsWithGATarget( final ProjectRef target,
                                                                              final ProjectRelationshipFilter filter )
        throws TensorDataException
    {
        return getAllDirectRelationshipsWithExactTarget( new ProjectVersionRef( target.getGroupId(),
                                                                                target.getArtifactId(), "*" ), filter );
    }

    //    @Override
    //    public EProjectRelationships getProjectRelationships( final EProjectKey key )
    //        throws TensorDataException
    //    {
    //        return store.getRelationshipSet( key );
    //    }
    //
    //    /* (non-Javadoc)
    //     * @see org.commonjava.tensor.data.TensorDataManager#getProjectRelationships(org.apache.maven.graph.common.ref.ProjectVersionRef)
    //     */
    //    @Override
    //    public EProjectRelationships getProjectRelationships( final ProjectVersionRef projectRef )
    //        throws TensorDataException
    //    {
    //        return store.getRelationshipSet( new EProjectKey( projectRef ) );
    //    }

    /* (non-Javadoc)
     * @see org.commonjava.tensor.data.TensorDataManager#contains(org.apache.maven.graph.common.ref.ProjectVersionRef)
     */
    @Override
    public boolean contains( final ProjectVersionRef ref )
    {
        return globalWeb.containsGraph( new EProjectKey( ref ) );
    }

    /* (non-Javadoc)
     * @see org.commonjava.tensor.data.TensorDataManager#getAllStoredProjectRefs()
     */
    @Override
    public Set<ProjectVersionRef> getAllStoredProjectRefs()
        throws TensorDataException
    {
        return globalWeb.getAllProjects();
    }

    @Override
    public Map<String, String> getMetadata( final EProjectKey key )
    {
        return globalWeb.getMetadata( key );
    }

    @Override
    public void addMetadata( final EProjectKey key, final String name, final String value )
    {
        globalWeb.addMetadata( key, name, value );
    }

    @Override
    public void addMetadata( final EProjectKey key, final Map<String, String> metadata )
    {
        globalWeb.addMetadata( key, metadata );
    }

    @Override
    public Set<ProjectVersionRef> getIncompleteSubgraphsFor( final ProjectVersionRef ref )
        throws TensorDataException
    {
        return getIncompleteSubgraphsFor( new EProjectKey( ref ) );
    }

    @Override
    public Set<ProjectVersionRef> getAllIncompleteSubgraphs()
        throws TensorDataException
    {
        return globalWeb.getIncompleteSubgraphs();
    }

    @Override
    public Set<ProjectVersionRef> getVariableSubgraphsFor( final ProjectVersionRef ref )
        throws TensorDataException
    {
        // TODO Auto-generated method stub
        return null;
    }

    @Override
    public Set<ProjectVersionRef> getAllVariableSubgraphs()
        throws TensorDataException
    {
        return globalWeb.getVariableSubgraphs();
    }

    @Override
    public Set<ProjectVersionRef> getIncompleteSubgraphsFor( final EProjectKey key )
        throws TensorDataException
    {
        try
        {
            final EProjectGraph graph = globalWeb.getGraph( key );

            return graph == null ? null : graph.getIncompleteSubgraphs();
        }
        catch ( final GraphDriverException e )
        {
            throw new TensorDataException( "Failed to retrieve graph for: %s from global graph database. Reason: %s",
                                           e, key.getProject(), e.getMessage() );
        }
    }

    @Override
    public Set<ProjectVersionRef> getVariableSubgraphsFor( final EProjectKey key )
        throws TensorDataException
    {
        try
        {
            final EProjectGraph graph = globalWeb.getGraph( key );

            return graph == null ? null : graph.getVariableSubgraphs();
        }
        catch ( final GraphDriverException e )
        {
            throw new TensorDataException( "Failed to retrieve graph for: %s from global graph database. Reason: %s",
                                           e, key.getProject(), e.getMessage() );
        }
    }

    @Override
    public boolean hasErrors( final ProjectVersionRef ref )
    {
        final Map<String, String> md = getMetadata( new EProjectKey( ref ) );
        if ( md == null )
        {
            return false;
        }

        return md.containsKey( MODEL_ERRORS );
    }

    @Override
    public Set<String> getErrors( final ProjectVersionRef ref )
    {
        final EProjectKey key = new EProjectKey( ref );

        final Map<String, String> md = getMetadata( key );
        if ( md == null )
        {
            return null;
        }

        final String serialized = md.get( MODEL_ERRORS );

        if ( isEmpty( serialized ) )
        {
            return null;
        }

        final String[] errors = serialized.split( ERROR_SEPARATOR );
        return new HashSet<String>( Arrays.asList( errors ) );
    }

    @Override
    public synchronized void addError( final ProjectVersionRef ref, final Throwable error )
    {
        final EProjectKey key = new EProjectKey( ref );

        final Map<String, String> md = getMetadata( key );
        if ( md == null )
        {
            return;
        }

        String serialized = md.get( MODEL_ERRORS );

        final String errorStr = toString( error );
        if ( isEmpty( serialized ) )
        {
            serialized = errorStr;
        }
        else
        {
            serialized += ERROR_SEPARATOR + errorStr;
        }

        addMetadata( key, MODEL_ERRORS, serialized );
    }

    private String toString( final Throwable e )
    {
        final StringWriter sw = new StringWriter();
        final PrintWriter pw = new PrintWriter( sw );

        e.printStackTrace( pw );

        return sw.toString();
    }

    @Override
    public Map<ProjectVersionRef, Set<String>> getAllProjectErrors()
        throws TensorDataException
    {
        final Set<ProjectVersionRef> projects = globalWeb.getProjectsWithMetadata( MODEL_ERRORS );
        final Map<ProjectVersionRef, Set<String>> errors =
            new HashMap<ProjectVersionRef, Set<String>>( projects.size() );

        for ( final ProjectVersionRef project : projects )
        {
            errors.put( project, getErrors( project ) );
        }

        return errors;
    }

    @Override
    public Map<ProjectVersionRef, Set<String>> getProjectErrorsInGraph( final ProjectVersionRef ref )
        throws TensorDataException
    {
        EProjectGraph graph;
        try
        {
            graph = globalWeb.getGraph( new EProjectKey( ref ) );
        }
        catch ( final GraphDriverException e )
        {
            throw new TensorDataException( "Failed to retrieve graph for: %s. Reason: %s", e, ref, e.getMessage() );
        }

        if ( graph == null )
        {
            return null;
        }

        final Set<ProjectVersionRef> projects = graph.getProjectsWithMetadata( MODEL_ERRORS );

        if ( projects == null )
        {
            return null;
        }

        final Map<ProjectVersionRef, Set<String>> errors =
            new HashMap<ProjectVersionRef, Set<String>>( projects.size() );

        for ( final ProjectVersionRef project : projects )
        {
            errors.put( project, getErrors( project ) );
        }

        return errors;
    }

    @Override
    public void reindex( final ProjectVersionRef ref )
        throws TensorDataException
    {
        EProjectGraph graph;
        try
        {
            graph = globalWeb.getGraph( new EProjectKey( ref ) );
        }
        catch ( final GraphDriverException e )
        {
            throw new TensorDataException( "Failed to retrieve graph: %s. Reason: %s", e, ref, e.getMessage() );
        }

        if ( graph == null )
        {
            throw new TensorDataException( "Graph not found: %s", ref );
        }

        try
        {
            graph.reindex();
        }
        catch ( final GraphDriverException e )
        {
            throw new TensorDataException( "Failed to reindex graph: %s. Reason: %s", e, ref, e.getMessage() );
        }
    }

    @Override
    public void reindexAll()
        throws TensorDataException
    {
        try
        {
            globalWeb.reindex();
        }
        catch ( final GraphDriverException e )
        {
            throw new TensorDataException( "Failed to reindex global graph. Reason: %s", e, e.getMessage() );
        }
    }
}
