package org.commonjava.tensor.discover;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.enterprise.context.ApplicationScoped;
import javax.enterprise.inject.Alternative;
import javax.inject.Inject;

import org.apache.http.HttpResponse;
import org.apache.http.StatusLine;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.ResponseHandler;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.maven.artifact.repository.metadata.Metadata;
import org.apache.maven.artifact.repository.metadata.Versioning;
import org.apache.maven.artifact.repository.metadata.io.xpp3.MetadataXpp3Reader;
import org.apache.maven.graph.common.ref.ProjectVersionRef;
import org.apache.maven.graph.common.version.InvalidVersionSpecificationException;
import org.apache.maven.graph.common.version.VersionSpec;
import org.apache.maven.graph.common.version.VersionUtils;
import org.apache.maven.model.InputSource;
import org.apache.maven.model.Model;
import org.apache.maven.model.io.xpp3.MavenXpp3ReaderEx;
import org.codehaus.plexus.util.xml.pull.XmlPullParserException;
import org.commonjava.tensor.data.TensorDataException;
import org.commonjava.tensor.util.MavenModelProcessor;
import org.commonjava.util.logging.Logger;

@ApplicationScoped
@Alternative
public class SimpleDiscoverer
    implements ProjectRelationshipDiscoverer
{

    private final Logger logger = new Logger( getClass() );

    private final DefaultHttpClient client;

    private final List<String> baseurls;

    @Inject
    private MavenModelProcessor modelProcessor;

    protected SimpleDiscoverer()
    {
        client = new DefaultHttpClient();
        this.baseurls = formatUrls( "http://repo.maven.apache.org/maven2" );
    }

    public SimpleDiscoverer( final MavenModelProcessor modelProcessor, final String... baseurls )
    {
        client = new DefaultHttpClient();
        this.baseurls = formatUrls( baseurls );
    }

    private List<String> formatUrls( final String... baseurls )
    {
        final List<String> urls = new ArrayList<String>();
        for ( final String url : baseurls )
        {
            if ( !url.endsWith( "/" ) )
            {
                urls.add( url + "/" );
            }
            else
            {
                urls.add( url );
            }
        }

        return Collections.unmodifiableList( urls );
    }

    @Override
    public ProjectVersionRef resolveSpecificVersion( final ProjectVersionRef ref )
        throws TensorDataException
    {
        if ( ref.isSnapshot() )
        {
            return resolveSnapshot( ref );
        }
        else if ( !ref.getVersionSpec()
                      .isSingle() )
        {
            return resolveRange( ref );
        }
        else
        {
            return ref;
        }
    }

    @Override
    public ProjectVersionRef discoverRelationships( final ProjectVersionRef ref, final DiscoveryConfig discoveryConfig )
        throws TensorDataException
    {
        ProjectVersionRef specific = ref;
        if ( !ref.isSpecificVersion() )
        {
            specific = resolveSpecificVersion( ref );
        }

        final StreamLocation location =
            getFirst( toGroupPath( specific.getGroupId(), specific.getArtifactId(), specific.getVersionSpec()
                                                                                            .renderStandard(),
                                   specific.getArtifactId() + "-" + specific.getVersionSpec()
                                                                            .renderStandard() + ".pom" ) );

        Model model = null;
        if ( location != null )
        {
            final InputSource is = new InputSource();
            is.setLocation( location.getLocation() );
            is.setModelId( specific.toString() );

            try
            {
                model = new MavenXpp3ReaderEx().read( location.getStream(), false, is );
            }
            catch ( final IOException e )
            {
                logger.error( "Failed to read POM for: '%s' from: '%s'. Reason: %s", e, specific,
                              location.getLocation(), e.getMessage() );
            }
            catch ( final XmlPullParserException e )
            {
                logger.error( "Failed to read POM for: '%s' from: '%s'. Reason: %s", e, specific,
                              location.getLocation(), e.getMessage() );
            }
        }

        if ( model != null )
        {
            modelProcessor.storeModelRelationships( model );
        }

        return specific;
    }

    private ProjectVersionRef resolveSnapshot( final ProjectVersionRef ref )
    {
        final StreamLocation location =
            getFirst( toGroupPath( ref.getGroupId(), ref.getArtifactId(), ref.getVersionSpec()
                                                                             .renderStandard(), "maven-metadata.xml" ) );
        String latest = null;
        try
        {
            final Metadata read = new MetadataXpp3Reader().read( location.getStream() );
            final Versioning versioning = read.getVersioning();

            latest = versioning.getLatest();
            if ( latest != null )
            {
                return new ProjectVersionRef( ref, latest );
            }
        }
        catch ( final IOException e )
        {
            logger.error( "Failed to parse snapshot metadata for: '%s'. Reason: %s", e, ref, e.getMessage() );
        }
        catch ( final XmlPullParserException e )
        {
            logger.error( "Failed to parse snapshot metadata for: '%s'. Reason: %s", e, ref, e.getMessage() );
        }
        catch ( final InvalidVersionSpecificationException e )
        {
            logger.error( "Latest version in snapshot metadata '%s' is cannot be parsed. Reason: %s", e, latest,
                          e.getMessage() );
        }

        return null;
    }

    private ProjectVersionRef resolveRange( final ProjectVersionRef ref )
    {
        final List<StreamLocation> locations =
            getAll( toGroupPath( ref.getGroupId(), ref.getArtifactId(), "maven-metadata.xml" ) );
        final List<String> allVersions = new ArrayList<String>();

        for ( final StreamLocation loc : locations )
        {
            try
            {
                final Metadata read = new MetadataXpp3Reader().read( loc.getStream() );
                final Versioning versioning = read.getVersioning();
                final List<String> versions = versioning.getVersions();

                if ( versions != null )
                {
                    for ( final String version : versions )
                    {
                        if ( !allVersions.contains( version ) )
                        {
                            allVersions.add( version );
                        }
                    }
                }
            }
            catch ( final IOException e )
            {
                logger.error( "Failed to parse artifact-level metadata for: '%s'. Reason: %s", e, ref, e.getMessage() );
            }
            catch ( final XmlPullParserException e )
            {
                logger.error( "Failed to parse artifact-level metadata for: '%s'. Reason: %s", e, ref, e.getMessage() );
            }
        }

        final List<VersionSpec> specs = new ArrayList<VersionSpec>();
        if ( allVersions != null && !allVersions.isEmpty() )
        {
            for ( final String spec : allVersions )
            {
                try
                {
                    specs.add( VersionUtils.createFromSpec( spec ) );
                }
                catch ( final InvalidVersionSpecificationException e )
                {
                    logger.error( "Unparsable version spec found in metadata: '%s'. Reason: %s", e, spec,
                                  e.getMessage() );
                }
            }
        }

        if ( !specs.isEmpty() )
        {
            Collections.sort( specs );
            return new ProjectVersionRef( ref, specs.get( specs.size() - 1 ) );
        }

        return null;
    }

    private List<StreamLocation> getAll( final String path )
    {
        final List<StreamLocation> all = new ArrayList<StreamLocation>();
        for ( final String base : baseurls )
        {
            final StreamLocation stream = getStream( base, path );
            if ( stream != null )
            {
                all.add( stream );
            }
        }

        return all;
    }

    private StreamLocation getFirst( final String path )
    {
        for ( final String base : baseurls )
        {
            final StreamLocation stream = getStream( base, path );

            if ( stream != null )
            {
                return stream;
            }
        }

        return null;
    }

    private StreamLocation getStream( final String base, final String path )
    {
        StreamLocation stream = null;
        if ( base.startsWith( "http" ) )
        {
            try
            {
                stream = getHttp( toPath( base, path ) );
            }
            catch ( final ClientProtocolException e )
            {
                logger.error( "Failed to retrieve path: '%s' from: '%s'. Reason: %s", e, path, base, e.getMessage() );
            }
            catch ( final IOException e )
            {
                logger.error( "Failed to retrieve path: '%s' from: '%s'. Reason: %s", e, path, base, e.getMessage() );
            }
        }
        else if ( base.startsWith( "file:" ) || base.startsWith( "/" ) )
        {
            final File f = new File( base, path );
            try
            {
                stream = new StreamLocation( f.getAbsolutePath(), new FileInputStream( f ) );
            }
            catch ( final FileNotFoundException e )
            {
                logger.error( "Failed to retrieve path: '%s' from: '%s'. Reason: %s", e, path, base, e.getMessage() );
            }
        }
        else
        {
            logger.error( "Invalid repository base: '%s'. Skipping..." );
        }

        return stream;
    }

    private StreamLocation getHttp( final String url )
        throws ClientProtocolException, IOException
    {
        final HttpGet get = new HttpGet( url );
        return client.execute( get, new StreamResponseHandler( new StreamLocation( url ) ) );
    }

    private String toGroupPath( final String groupId, final String... parts )
    {
        return toPath( groupId.replace( '.', '/' ), parts );
    }

    private String toPath( final String base, final String... parts )
    {
        final StringBuilder sb = new StringBuilder( base );
        for ( final String part : parts )
        {
            if ( sb.charAt( sb.length() - 1 ) != '/' && part.charAt( 0 ) != '/' )
            {
                sb.append( '/' );
            }

            sb.append( part );
        }

        return sb.toString();
    }

    private static final class StreamResponseHandler
        implements ResponseHandler<StreamLocation>
    {
        private final StreamLocation streamLocation;

        public StreamResponseHandler( final StreamLocation streamLocation )
        {
            this.streamLocation = streamLocation;
        }

        @Override
        public StreamLocation handleResponse( final HttpResponse response )
            throws ClientProtocolException, IOException
        {
            final StatusLine sl = response.getStatusLine();
            if ( sl.getStatusCode() == 200 )
            {
                streamLocation.setStream( response.getEntity()
                                                  .getContent() );
                return streamLocation;
            }
            return null;
        }

    }

    public class StreamLocation
    {

        private final String location;

        private InputStream stream;

        public StreamLocation( final String location, final InputStream stream )
        {
            this.location = location;
            this.stream = stream;
        }

        public StreamLocation( final String location )
        {
            this.location = location;
        }

        public void setStream( final InputStream stream )
        {
            this.stream = stream;
        }

        public String getLocation()
        {
            return location;
        }

        public InputStream getStream()
        {
            return stream;
        }
    }

}
