/*******************************************************************************
 * Copyright 2011 John Casey
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.commonjava.tensor.util;

import static org.apache.commons.lang.StringUtils.isEmpty;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.inject.Inject;

import org.apache.maven.graph.common.DependencyScope;
import org.apache.maven.graph.common.ref.ArtifactRef;
import org.apache.maven.graph.common.ref.ProjectVersionRef;
import org.apache.maven.graph.common.version.InvalidVersionSpecificationException;
import org.apache.maven.graph.effective.EProjectRelationships;
import org.apache.maven.graph.effective.EProjectRelationships.Builder;
import org.apache.maven.graph.effective.rel.DependencyRelationship;
import org.apache.maven.graph.effective.rel.ExtensionRelationship;
import org.apache.maven.graph.effective.rel.PluginRelationship;
import org.apache.maven.model.Build;
import org.apache.maven.model.Dependency;
import org.apache.maven.model.DependencyManagement;
import org.apache.maven.model.Extension;
import org.apache.maven.model.Model;
import org.apache.maven.model.Parent;
import org.apache.maven.model.Plugin;
import org.apache.maven.model.PluginManagement;
import org.apache.maven.model.ReportPlugin;
import org.apache.maven.model.Reporting;
import org.cdmckay.coffeedom.CoffeeDOMException;
import org.cdmckay.coffeedom.Document;
import org.cdmckay.coffeedom.Element;
import org.cdmckay.coffeedom.input.SAXBuilder;
import org.cdmckay.coffeedom.xpath.XPath;
import org.codehaus.plexus.interpolation.InterpolationException;
import org.codehaus.plexus.interpolation.PrefixedObjectValueSource;
import org.codehaus.plexus.interpolation.PropertiesBasedValueSource;
import org.codehaus.plexus.interpolation.StringSearchInterpolator;
import org.commonjava.tensor.data.TensorDataException;
import org.commonjava.tensor.data.TensorDataManager;
import org.commonjava.util.logging.Logger;

@javax.enterprise.context.ApplicationScoped
public class MavenModelProcessor
{

    private static final String SITE_PLUGIN = "org.apache.maven.plugins:maven-site-plugin";

    private static final Logger logger = new Logger( MavenModelProcessor.class );

    @Inject
    private TensorDataManager dataManager;

    protected MavenModelProcessor()
    {
    }

    public MavenModelProcessor( final TensorDataManager dataManager )
    {
        this.dataManager = dataManager;
    }

    public EProjectRelationships storeModelRelationships( final Model model )
        throws TensorDataException
    {
        try
        {
            String g = model.getGroupId();
            String v = model.getVersion();

            final Parent parent = model.getParent();
            if ( parent != null )
            {
                if ( g == null )
                {
                    g = parent.getGroupId();
                }

                if ( v == null )
                {
                    v = parent.getVersion();
                }
            }

            v = resolveExpressions( v, model );
            g = resolveExpressions( g, model );

            String a = model.getArtifactId();

            a = resolveExpressions( a, model );

            final ProjectVersionRef projectRef = new ProjectVersionRef( g, a, v );

            final EProjectRelationships.Builder builder = new EProjectRelationships.Builder( projectRef );

            addParentRelationship( builder, model, projectRef );
            addDependencyRelationships( builder, model, projectRef );
            addPluginUsages( builder, model, projectRef );
            addExtensionUsages( builder, model, projectRef );

            final EProjectRelationships rels = builder.build();
            dataManager.storeRelationships( rels );

            return rels;
        }
        catch ( final TensorDataException e )
        {
            throw e;
        }
        catch ( final InvalidVersionSpecificationException e )
        {
            throw new TensorDataException( "Failed to parse version string: '%s' for model: %s. Reason: %s", e,
                                           model.getVersion(), model, e.getMessage() );
        }
    }

    private String resolveExpressions( final String raw, final Model model )
        throws TensorDataException
    {
        if ( raw.contains( "${" ) )
        {
            final StringSearchInterpolator interp = new StringSearchInterpolator();
            interp.addValueSource( new PropertiesBasedValueSource( model.getProperties() ) );

            final List<String> expressionRoots = new ArrayList<String>();
            expressionRoots.add( "pom." );
            expressionRoots.add( "project." );

            interp.addValueSource( new PrefixedObjectValueSource( expressionRoots, model, true ) );

            try
            {
                return interp.interpolate( raw );
            }
            catch ( final InterpolationException e )
            {
                throw new TensorDataException(
                                               "Failed to resolve expression from model.\nRaw string: '%s'\nModel: %s\nError: %s",
                                               e, raw, model, e.getMessage() );
            }
        }

        return raw;
    }

    private void addExtensionUsages( final Builder builder, final Model model, final ProjectVersionRef projectRef )
        throws TensorDataException
    {
        final List<Extension> extensions =
            model.getBuild() == null ? new ArrayList<Extension>() : model.getBuild()
                                                                         .getExtensions();

        for ( final Extension ext : extensions )
        {
            if ( ext == null )
            {
                continue;
            }
            else if ( isEmpty( ext.getVersion() ) )
            {
                // TODO: Do we resolve LATEST here? I thought this was an invalid state??
                continue;
            }

            final String g = resolveExpressions( ext.getGroupId(), model );
            final String a = resolveExpressions( ext.getArtifactId(), model );
            final String v = resolveExpressions( ext.getVersion(), model );
            final ProjectVersionRef ref = new ProjectVersionRef( g, a, v );

            //            if ( isValid( ref ) )
            //            {
            builder.withExtensions( new ExtensionRelationship( projectRef, ref, builder.getNextExtensionIndex() ) );
            //            }
        }
    }

    public void addPluginUsages( final Builder builder, final Model model, final ProjectVersionRef projectRef )
        throws TensorDataException
    {
        addPluginUsages( builder, model, false, projectRef );
        addPluginUsages( builder, model, true, projectRef );
        addReportPluginUsages( builder, model, projectRef );
        addSiteReportPluginUsages( builder, model, projectRef );
    }

    private void addSiteReportPluginUsages( final Builder builder, final Model model, final ProjectVersionRef projectRef )
        throws TensorDataException
    {
        final Build build = model.getBuild();
        if ( build != null )
        {
            List<Plugin> plugins = build.getPlugins();
            for ( final Plugin plugin : plugins )
            {
                if ( plugin.getKey()
                           .equals( SITE_PLUGIN ) )
                {
                    addSiteReportPlugins( builder, plugin, plugin.getConfiguration(), projectRef, model );
                }
            }

            final PluginManagement pmgmt = build.getPluginManagement();
            if ( pmgmt != null )
            {
                plugins = pmgmt.getPlugins();
                for ( final Plugin plugin : plugins )
                {
                    if ( plugin.getKey()
                               .equals( SITE_PLUGIN ) )
                    {
                        addSiteReportPlugins( builder, plugin, plugin.getConfiguration(), projectRef, model );
                    }
                }
            }
        }
    }

    public void addSiteReportPlugins( final Builder builder, final Plugin plugin, final Object configuration,
                                      final ProjectVersionRef projectRef, final Model model )
        throws TensorDataException
    {
        XPath reportPluginPath = null;
        try
        {
            reportPluginPath = XPath.newInstance( "configuration/reportPlugins/*" );
        }
        catch ( final CoffeeDOMException e )
        {
            logger.error( "Cannot compile report-plugin selection XPath for site-plugin configuration. Error: "
                              + e.getMessage(), e );
        }

        if ( reportPluginPath == null || configuration == null )
        {
            return;
        }

        final String xml = configuration.toString();
        try
        {
            final Document configDoc = new SAXBuilder().build( new ByteArrayInputStream( xml.getBytes() ) );

            final List<Object> pluginNodes = reportPluginPath.selectNodes( configDoc );

            if ( pluginNodes != null )
            {
                for ( final Object node : pluginNodes )
                {
                    final Element pluginNode = (Element) node;
                    final Element gEl = pluginNode.getChild( "groupId" );
                    final Element aEl = pluginNode.getChild( "artifactId" );
                    final Element vEl = pluginNode.getChild( "version" );

                    if ( gEl == null || aEl == null )
                    {
                        logger.debug( "Cannot process invalid report-plugin configuration: %s:%s", gEl, aEl );
                        continue;
                    }

                    String gid = gEl.getTextNormalize();
                    String aid = aEl.getTextNormalize();

                    String ver = null;
                    if ( vEl != null )
                    {
                        ver = vEl.getTextNormalize();
                    }

                    if ( isEmpty( ver ) )
                    {
                        //TODO: Resolve LATEST at this point in time...?
                        logger.debug( "No version found for: %s:%s. Skipping.", gEl, aEl );
                        continue;
                    }

                    gid = resolveExpressions( gid, model );
                    aid = resolveExpressions( aid, model );
                    ver = resolveExpressions( ver, model );
                    final ProjectVersionRef ref = new ProjectVersionRef( gid, aid, ver );

                    //                    if ( isValid( ref ) )
                    //                    {
                    builder.withPlugins( new PluginRelationship( projectRef, ref, builder.getNextPluginIndex( false ),
                                                                 false ) );
                    //                    }
                }
            }
        }
        catch ( final CoffeeDOMException e )
        {
            logger.error( "Cannot select report plugin definitions from site-plugin configuration. Error: "
                              + e.getMessage(), e );
        }
        catch ( final IOException e )
        {
            logger.error( "Cannot read site-plugin configuration. Error: " + e.getMessage(), e );
        }
    }

    public void addReportPluginUsages( final Builder builder, final Model model, final ProjectVersionRef projectRef )
        throws TensorDataException
    {
        final Reporting reporting = model.getReporting();
        if ( reporting != null )
        {
            final List<ReportPlugin> plugins = reporting.getPlugins();
            for ( final ReportPlugin plugin : plugins )
            {
                if ( plugin == null )
                {
                    continue;
                }
                else if ( isEmpty( plugin.getVersion() ) )
                {
                    //TODO: Resolve LATEST at this point in time...?
                    continue;
                }

                final String g = resolveExpressions( plugin.getGroupId(), model );
                final String a = resolveExpressions( plugin.getArtifactId(), model );
                final String v = resolveExpressions( plugin.getVersion(), model );
                final ProjectVersionRef ref = new ProjectVersionRef( g, a, v );

                //                if ( isValid( ref ) )
                //                {
                builder.withPlugins( new PluginRelationship( projectRef, ref,
                                                             builder.getNextPluginDependencyIndex( projectRef, false ),
                                                             false ) );
                //                }
            }
        }
    }

    public void addPluginUsages( final Builder builder, final Model model, final boolean managed,
                                 final ProjectVersionRef projectRef )
        throws TensorDataException
    {
        final List<Plugin> plugins = getPlugins( model, managed );

        for ( final Plugin plugin : plugins )
        {
            if ( plugin == null )
            {
                continue;
            }
            else if ( isEmpty( plugin.getVersion() ) )
            {
                //TODO: Resolve LATEST at this point in time...?
                continue;
            }

            final String g = resolveExpressions( plugin.getGroupId(), model );
            final String a = resolveExpressions( plugin.getArtifactId(), model );
            final String v = resolveExpressions( plugin.getVersion(), model );
            final ProjectVersionRef ref = new ProjectVersionRef( g, a, v );

            //            if ( isValid( ref ) )
            //            {
            builder.withPlugins( new PluginRelationship( projectRef, ref,
                                                         builder.getNextPluginDependencyIndex( projectRef, managed ),
                                                         managed ) );
            //            }
        }
    }

    private List<Plugin> getPlugins( final Model model, final boolean managed )
    {
        final Build build = model.getBuild();
        if ( build == null )
        {
            return Collections.emptyList();
        }

        List<Plugin> plugins = null;
        if ( managed )
        {
            final PluginManagement pmgmt = build.getPluginManagement();
            if ( pmgmt != null )
            {
                plugins = pmgmt.getPlugins();
            }
        }
        else
        {
            plugins = build.getPlugins();
        }

        if ( plugins == null )
        {
            return Collections.emptyList();
        }

        return plugins;
    }

    public void addDependencyRelationships( final Builder builder, final Model model, final ProjectVersionRef projectRef )
        throws TensorDataException, InvalidVersionSpecificationException
    {
        addDependencyRelationships( builder, model, true, projectRef );
        addDependencyRelationships( builder, model, false, projectRef );
    }

    public void addDependencyRelationships( final Builder builder, final Model model, final boolean managed,
                                            final ProjectVersionRef projectRef )
        throws TensorDataException
    {
        if ( managed )
        {
            final DependencyManagement dm = model.getDependencyManagement();
            if ( dm != null )
            {
                for ( final Dependency dep : dm.getDependencies() )
                {
                    final String g = resolveExpressions( dep.getGroupId(), model );
                    final String a = resolveExpressions( dep.getArtifactId(), model );
                    final String v = resolveExpressions( dep.getVersion(), model );
                    final ProjectVersionRef depRef = new ProjectVersionRef( g, a, v );

                    final ArtifactRef artifactRef =
                        new ArtifactRef( depRef, dep.getType(), dep.getClassifier(), dep.isOptional() );
                    //                    if ( isValid( artifactRef ) )
                    //                    {
                    builder.withDependencies( new DependencyRelationship( projectRef, artifactRef,
                                                                          DependencyScope.getScope( dep.getScope() ),
                                                                          builder.getNextDependencyIndex( managed ),
                                                                          managed ) );
                    //                    }
                }
            }
        }
        else
        {
            for ( final Dependency dep : model.getDependencies() )
            {
                final String g = resolveExpressions( dep.getGroupId(), model );
                final String a = resolveExpressions( dep.getArtifactId(), model );
                final String v = resolveExpressions( dep.getVersion(), model );
                final ProjectVersionRef depRef = new ProjectVersionRef( g, a, v );

                final ArtifactRef artifactRef =
                    new ArtifactRef( depRef, dep.getType(), dep.getClassifier(), dep.isOptional() );
                //                if ( isValid( artifactRef ) )
                //                {
                builder.withDependencies( new DependencyRelationship( projectRef, artifactRef,
                                                                      DependencyScope.getScope( dep.getScope() ),
                                                                      builder.getNextDependencyIndex( managed ),
                                                                      managed ) );
                //                }
            }
        }
    }

    public void addParentRelationship( final Builder builder, final Model model, final ProjectVersionRef projectRef )
        throws TensorDataException
    {
        final Parent parent = model.getParent();
        if ( parent != null )
        {
            final String g = resolveExpressions( parent.getGroupId(), model );
            final String a = resolveExpressions( parent.getArtifactId(), model );
            final String v = resolveExpressions( parent.getVersion(), model );
            final ProjectVersionRef ref = new ProjectVersionRef( g, a, v );
            //            if ( isValid( ref ) )
            //            {
            builder.withParent( ref );
            //            }
        }
    }

    //    private boolean isValid( final ProjectVersionRef ref )
    //    {
    //        boolean valid = true;
    //        try
    //        {
    //            ref.getVersionSpec();
    //        }
    //        catch ( final InvalidVersionSpecificationException e )
    //        {
    //            logger.error( "Invalid version: %s. Reason: %s", e, ref, e.getMessage() );
    //            valid = false;
    //        }
    //
    //        return valid;
    //    }
}
